
package org.act.services.thread;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.logging.Logger;
import org.act.services.locale.StringManager;

public class ThreadPool {
	private static Logger log = Logger.getLogger(ThreadPool.class.getName());
	private static StringManager sm =
		StringManager.getManager("org.act.services.locale");
	private static boolean logfull = true;
	
	public static final int MAX_THREADS = 200;
	public static final int MAX_THREADS_MIN = 10;
	public static final int MAX_SPARE_THREADS = 50;
	public static final int MIN_SPARE_THREADS = 4;
	public static final int WORK_WAIT_TIMEOUT = 60 * 1000;
	
	protected ControlRunnable[] pool = null;
	
	protected MonitorRunnable monitor;
	
	protected int maxThreads;
	
	protected int minSpareThreads;
	
	protected int maxSpareThreads;
	
	protected int currentThreadCount;
	
	protected int currentThreadsBusy;
	
	protected boolean stopThePool;
	
	protected boolean isDaemon = true;
	
	protected Hashtable threads = new Hashtable();
	protected Vector listeners = new Vector();
	
	protected String name = "TP";
	
	protected int sequence = 1;
	
	protected int threadPriority = Thread.NORM_PRIORITY;
	
	public ThreadPool() {
		maxThreads = MAX_THREADS;
		maxSpareThreads = MAX_SPARE_THREADS;
		minSpareThreads = MIN_SPARE_THREADS;
		currentThreadCount = 0;
		currentThreadsBusy = 0;
		stopThePool = false;
	}
	
	public static ThreadPool createThreadPool(boolean jmx) {
		return new ThreadPool();
	}
	public synchronized void start() {
		stopThePool = false;
		currentThreadCount = 0;
		currentThreadsBusy = 0;
		adjustLimits();
		pool = new ControlRunnable[maxThreads];
		openThreads(minSpareThreads);
		if (maxSpareThreads < maxThreads) {
			monitor = new MonitorRunnable(this);
		}
	}
	public MonitorRunnable getMonitor() {
		return monitor;
	}
	
	public synchronized void setThreadPriority(int threadPriority) {
								if (threadPriority < Thread.MIN_PRIORITY) {
			throw new IllegalArgumentException("new priority < MIN_PRIORITY");
		} else if (threadPriority > Thread.MAX_PRIORITY) {
			throw new IllegalArgumentException("new priority > MIN_PRIORITY");
		}
				this.threadPriority = threadPriority;
		Enumeration currentThreads = getThreads();
		Thread t = null;
		while (currentThreads.hasMoreElements()) {
			t = (Thread) currentThreads.nextElement();
			t.setPriority(threadPriority);
		}
	}
	
	public int getThreadPriority() {
		return threadPriority;
	}
	public void setMaxThreads(int maxThreads) {
		this.maxThreads = maxThreads;
	}
	public int getMaxThreads() {
		return maxThreads;
	}
	public void setMinSpareThreads(int minSpareThreads) {
		this.minSpareThreads = minSpareThreads;
	}
	public int getMinSpareThreads() {
		return minSpareThreads;
	}
	public void setMaxSpareThreads(int maxSpareThreads) {
		this.maxSpareThreads = maxSpareThreads;
	}
	public int getMaxSpareThreads() {
		return maxSpareThreads;
	}
	public int getCurrentThreadCount() {
		return currentThreadCount;
	}
	public int getCurrentThreadsBusy() {
		return currentThreadsBusy;
	}
	public boolean isDaemon() {
		return isDaemon;
	}
	public static int getDebug() {
		return 0;
	}
	
	public void setDaemon(boolean b) {
		isDaemon = b;
	}
	public boolean getDaemon() {
		return isDaemon;
	}
	public void setName(String name) {
		this.name = name;
	}
	public String getName() {
		return name;
	}
	public int getSequence() {
		return sequence++;
	}
	public void addThread(Thread t, ControlRunnable cr) {
		threads.put(t, cr);
		for (int i = 0; i < listeners.size(); i++) {
			ThreadPoolListener tpl =
				(ThreadPoolListener) listeners.elementAt(i);
			tpl.threadStart(this, t);
		}
	}
	public void removeThread(Thread t) {
		threads.remove(t);
		for (int i = 0; i < listeners.size(); i++) {
			ThreadPoolListener tpl =
				(ThreadPoolListener) listeners.elementAt(i);
			tpl.threadEnd(this, t);
		}
	}
	public void addThreadPoolListener(ThreadPoolListener tpl) {
		listeners.addElement(tpl);
	}
	public Enumeration getThreads() {
		return threads.keys();
	}
	public void run(Runnable r) {
		ControlRunnable c = findControlRunnable();
		c.runIt(r);
	}
							
	public void runIt(ThreadPoolRunnable r) {
		if (null == r) {
			throw new NullPointerException();
		}
		ControlRunnable c = findControlRunnable();
		c.runIt(r);
	}
	private ControlRunnable findControlRunnable() {
		ControlRunnable c = null;
		if (stopThePool) {
			throw new IllegalStateException();
		}
				synchronized (this) {
			while (currentThreadsBusy == currentThreadCount) {
								if (currentThreadCount < maxThreads) {
															int toOpen = currentThreadCount + minSpareThreads;
					openThreads(toOpen);
				} else {
															try {
						this.wait();
					}
																														catch (InterruptedException e) {
											}
																														if (stopThePool) {
						break;
					}
				}
			}
						if (0 == currentThreadCount || stopThePool) {
				throw new IllegalStateException();
			}
						int pos = currentThreadCount - currentThreadsBusy - 1;
			c = pool[pos];
			pool[pos] = null;
			currentThreadsBusy++;
		}
		return c;
	}
													
	public synchronized void shutdown() {
		if (!stopThePool) {
			stopThePool = true;
			if (monitor != null) {
				monitor.terminate();
				monitor = null;
			}
			for (int i = 0; i < currentThreadCount - currentThreadsBusy; i++) {
				try {
					pool[i].terminate();
				} catch (Throwable t) {
					
									}
			}
			currentThreadsBusy = currentThreadCount = 0;
			pool = null;
			notifyAll();
		}
	}
	
	protected synchronized void checkSpareControllers() {
		if (stopThePool) {
			return;
		}
		if ((currentThreadCount - currentThreadsBusy) > maxSpareThreads) {
			int toFree =
				currentThreadCount - currentThreadsBusy - maxSpareThreads;
			for (int i = 0; i < toFree; i++) {
				ControlRunnable c =
					pool[currentThreadCount - currentThreadsBusy - 1];
				c.terminate();
				pool[currentThreadCount - currentThreadsBusy - 1] = null;
				currentThreadCount--;
			}
		}
	}
	
	protected synchronized void returnController(ControlRunnable c) {
		if (0 == currentThreadCount || stopThePool) {
			c.terminate();
			return;
		}
				currentThreadsBusy--;
		pool[currentThreadCount - currentThreadsBusy - 1] = c;
		notify();
	}
	
	protected synchronized void notifyThreadEnd(ControlRunnable c) {
		currentThreadsBusy--;
		currentThreadCount--;
		notify();
	}
	
	protected void adjustLimits() {
		if (maxThreads <= 0) {
			maxThreads = MAX_THREADS;
		} else if (maxThreads < MAX_THREADS_MIN) {
												maxThreads = MAX_THREADS_MIN;
		}
		if (maxSpareThreads >= maxThreads) {
			maxSpareThreads = maxThreads;
		}
		if (maxSpareThreads <= 0) {
			if (1 == maxThreads) {
				maxSpareThreads = 1;
			} else {
				maxSpareThreads = maxThreads / 2;
			}
		}
		if (minSpareThreads > maxSpareThreads) {
			minSpareThreads = maxSpareThreads;
		}
		if (minSpareThreads <= 0) {
			if (1 == maxSpareThreads) {
				minSpareThreads = 1;
			} else {
				minSpareThreads = maxSpareThreads / 2;
			}
		}
	}
	
	protected void openThreads(int toOpen) {
		if (toOpen > maxThreads) {
			toOpen = maxThreads;
		}
		for (int i = currentThreadCount; i < toOpen; i++) {
			pool[i - currentThreadsBusy] = new ControlRunnable(this);
		}
		currentThreadCount = toOpen;
	}
	
	void log(String s) {
		log.info(s);
			}
	
	public static class MonitorRunnable implements Runnable {
		ThreadPool p;
		Thread t;
		int interval = WORK_WAIT_TIMEOUT;
		boolean shouldTerminate;
		MonitorRunnable(ThreadPool p) {
			this.p = p;
			this.start();
		}
		public void start() {
			shouldTerminate = false;
			t = new Thread(this);
			t.setDaemon(p.getDaemon());
			t.setName(p.getName() + "-Monitor");
			t.start();
		}
		public void setInterval(int i) {
			this.interval = i;
		}
		public void run() {
			while (true) {
				try {
										synchronized (this) {
						this.wait(interval);
					}
															if (shouldTerminate) {
						break;
					}
										p.checkSpareControllers();
				} catch (Throwable t) {
									}
			}
		}
		public void stop() {
			this.terminate();
		}
		
		public synchronized void terminate() {
			shouldTerminate = true;
			this.notify();
		}
	}
	
	public static class ControlRunnable implements Runnable {
		
		private ThreadPool p;
		
		private ThreadWithAttributes t;
		
		private ThreadPoolRunnable toRun;
		private Runnable toRunRunnable;
		
		private boolean shouldTerminate;
		
		private boolean shouldRun;
		
		private boolean noThData;
		
		ControlRunnable(ThreadPool p) {
			toRun = null;
			shouldTerminate = false;
			shouldRun = false;
			this.p = p;
			t = new ThreadWithAttributes(p, this);
			t.setDaemon(true);
			t.setName(p.getName() + "-Processor" + p.getSequence());
			t.setPriority(p.getThreadPriority());
			p.addThread(t, this);
			noThData = true;
			t.start();
		}
		public void run() {
			boolean _shouldRun = false;
			boolean _shouldTerminate = false;
			ThreadPoolRunnable _toRun = null;
			try {
				while (true) {
					try {
						
						synchronized (this) {
							while (!shouldRun && !shouldTerminate) {
								this.wait();
							}
							_shouldRun = shouldRun;
							_shouldTerminate = shouldTerminate;
							_toRun = toRun;
						}
						if (_shouldTerminate) {
																					break;
						}
						
						try {
							if (noThData) {
								if (_toRun != null) {
									Object thData[] = _toRun.getInitData();
									t.setThreadData(p, thData);
																																			}
								noThData = false;
							}
							if (_shouldRun) {
								if (_toRun != null) {
									_toRun.runIt(t.getThreadData(p));
								} else if (toRunRunnable != null) {
									toRunRunnable.run();
								} else {
																										}
							}
						} catch (Throwable t) {
																					
														t.printStackTrace();
							shouldTerminate = true;
							shouldRun = false;
							p.notifyThreadEnd(this);
						} finally {
							if (_shouldRun) {
								shouldRun = false;
								
								p.returnController(this);
							}
						}
						
						if (_shouldTerminate) {
							break;
						}
					} catch (InterruptedException ie) { 
																	}
				}
			} finally {
				p.removeThread(Thread.currentThread());
			}
		}
		
		public synchronized void runIt(Runnable toRun) {
			this.toRunRunnable = toRun;
															shouldRun = true;
			this.notify();
		}
		
		public synchronized void runIt(ThreadPoolRunnable toRun) {
			this.toRun = toRun;
															shouldRun = true;
			this.notify();
		}
		public void stop() {
			this.terminate();
		}
		public synchronized void terminate() {
			shouldTerminate = true;
			this.notify();
		}
	}
	
	public String threadStatusString() {
		StringBuffer sb = new StringBuffer();
		Iterator it = threads.keySet().iterator();
		sb.append("<ul>");
		while (it.hasNext()) {
			sb.append("<li>");
			ThreadWithAttributes twa = (ThreadWithAttributes) it.next();
			sb.append(twa.getCurrentStage(this)).append(" ");
			sb.append(twa.getParam(this));
			sb.append("</li>\n");
		}
		sb.append("</ul>");
		return sb.toString();
	}
	
	public String[] getThreadStatus() {
		String status[] = new String[threads.size()];
		Iterator it = threads.keySet().iterator();
		for (int i = 0;(i < status.length && it.hasNext()); i++) {
			ThreadWithAttributes twa = (ThreadWithAttributes) it.next();
			status[i] = twa.getCurrentStage(this);
		}
		return status;
	}
	
	public String[] getThreadParam() {
		String status[] = new String[threads.size()];
		Iterator it = threads.keySet().iterator();
		for (int i = 0;(i < status.length && it.hasNext()); i++) {
			ThreadWithAttributes twa = (ThreadWithAttributes) it.next();
			Object o = twa.getParam(this);
			status[i] = (o == null) ? null : o.toString();
		}
		return status;
	}
	
	public static interface ThreadPoolListener {
		public void threadStart(ThreadPool tp, Thread t);
		public void threadEnd(ThreadPool tp, Thread t);
	}
}
